use super::Error;
use std::{ops::Deref, str::FromStr};

#[derive(Clone, PartialEq, Eq, Debug, Copy)]
/// Expression operators
///
/// The operators are used to compare values in the expression language.
pub enum ExprOp {
    And,
    Or,
    Eq,
    NotEq,
    Gt,
    Gte,
    Not,
    Lt,
    Lte,
    Add,
    Sub,
    Mul,
    Div,
}

impl From<&str> for ExprOp {
    fn from(s: &str) -> Self {
        match s.to_lowercase().as_str() {
            "+" => ExprOp::Add,
            "-" => ExprOp::Sub,
            "*" => ExprOp::Mul,
            "!" => ExprOp::Not,
            "/" => ExprOp::Div,
            "&&" | "and" => ExprOp::And,
            "||" | "or" => ExprOp::Or,
            "==" | "=" => ExprOp::Eq,
            "!=" => ExprOp::NotEq,
            ">" => ExprOp::Gt,
            ">=" => ExprOp::Gte,
            "<" => ExprOp::Lt,
            "<=" => ExprOp::Lte,
            x => panic!("{} is not a valid Operation", x),
        }
    }
}

#[derive(Clone, PartialEq, Eq, Debug, Copy)]
/// Order
///
/// The order is used to sort the results of a query.
pub enum Order {
    Ascending,
    Descending,
}

impl FromStr for Order {
    type Err = Error;

    fn from_str(s: &str) -> Result<Self, Self::Err> {
        match s.to_lowercase().as_str() {
            "asc" => Ok(Order::Ascending),
            "desc" => Ok(Order::Descending),
            s => Err(Error::InvalidOrder(s.to_owned())),
        }
    }
}

#[derive(Debug, Clone, PartialEq, Eq)]
/// Expression
///
/// This is the main expression type that is used to represent the AST of the expression.
///
/// It can be a simple variable, a string, a number, an identifier, a unary operation or a binary
/// operation.
pub enum Expr {
    Op { op: ExprOp, terms: Vec<Box<Expr>> },
    Variable(Variable),
    String(String),
    Number(i128),
}

#[derive(Clone, Debug, PartialEq, Hash, PartialOrd, Eq)]
/// Variable
///
/// A variable is a list of strings that represent a path to a value in the data.
pub struct Variable(Vec<String>);

impl Deref for Variable {
    type Target = [String];
    fn deref(&self) -> &Self::Target {
        &self.0
    }
}

impl From<Vec<String>> for Variable {
    fn from(v: Vec<String>) -> Self {
        Variable(v)
    }
}

impl From<&str> for Variable {
    fn from(s: &str) -> Self {
        Variable(vec![s.to_owned()])
    }
}

impl Variable {
    pub fn add_part(&mut self, part: String) {
        self.0.push(part);
    }
}
