use super::{
    expr::Variable,
    opcode::OpCode,
    value::{Value, ValueOrRef},
    Addr, Error, Register,
};
use crate::Transaction;
use num::CheckedAdd;
use std::{
    collections::{HashMap, VecDeque},
    ops::Deref,
};

macro_rules! get {
    ($r:expr,$pos:expr) => {
        ($r.get($pos)
            .ok_or_else(|| Error::RegisterOutOfBoundaries($pos.to_owned()))?)
    };
}

macro_rules! remove {
    ($r:expr,$pos:expr) => {
        ($r.remove($pos)
            .ok_or_else(|| Error::RegisterOutOfBoundaries($pos.to_owned()))?)
    };
}

macro_rules! set {
    ($r:expr, $pos:expr, $new_value:expr) => {
        (if let Some(previous_value) = $r.get_mut($pos) {
            *previous_value = $new_value;
        } else {
            $r.insert(($pos).to_owned(), $new_value);
        })
    };
}

#[inline]
pub fn execute<'a>(
    external_variables: &'a HashMap<Variable, ValueOrRef<'a>>,
    code: &'a [OpCode],
    initial_registers: Vec<Value>,
    start_at: Addr,
) -> Result<Value, Error> {
    let mut execution = start_at;
    let mut registers = initial_registers
        .iter()
        .enumerate()
        .map(|(pos, a)| (pos.into(), a.into()))
        .collect::<HashMap<Register, ValueOrRef>>();

    loop {
        match code.get(*execution).ok_or(Error::OutOfBoundaries)? {
            OpCode::LOAD_EXTERNAL(dst, name) => {
                let value = external_variables
                    .get(name)
                    .ok_or(Error::VariableNotFound((*name).join(".")))?;

                set!(registers, dst, value.clone());
            }
            OpCode::LOAD(dst, ref val) => set!(registers, dst, val.into()),
            OpCode::CPY(dst, reg2) => {
                let value = get!(registers, reg2).clone();
                set!(registers, dst, value);
            }
            OpCode::MOV(dst, reg2) => {
                let previous_value = remove!(registers, reg2);
                set!(registers, dst, previous_value);
            }
            OpCode::ADD(dst, a, b) => {
                let new_value = get!(registers, a)
                    .checked_add(get!(registers, b))
                    .ok_or(Error::Overflow)?
                    .into();
                set!(registers, dst, new_value);
            }
            OpCode::AND(dst, reg2, reg3) => {
                let new_value = Value::Bool(
                    get!(registers, reg2).as_boolean()? && get!(registers, reg3).as_boolean()?,
                )
                .into();
                set!(registers, dst, new_value);
            }
            OpCode::EQ(dst, reg1, reg2) => {
                let new_value = Value::Bool(get!(registers, reg1) == get!(registers, reg2)).into();
                set!(registers, dst, new_value);
            }
            OpCode::MUL(dst, reg1, reg2) => {
                let num1 = get!(registers, reg1).as_number()?;
                let num2 = get!(registers, reg2).as_number()?;
                set!(
                    registers,
                    dst,
                    num1.checked_mul(num2)
                        .map(|number| Value::Number(number).into())
                        .ok_or(Error::Overflow)?
                );
            }
            OpCode::OR(dst1, reg2, reg3) => {
                todo!()
            }
            OpCode::XOR(dst, reg2, reg3) => {
                todo!()
            }
            OpCode::NOT(dst, reg2) => {
                let new_value = Value::Bool(!get!(registers, reg2).as_boolean()?).into();
                set!(registers, dst, new_value);
            }
            OpCode::JMP(addr) => {
                execution.jump_to(addr);
                continue;
            }
            OpCode::JEQ(reg, addr) => {
                if get!(registers, reg).as_boolean()? {
                    execution.jump_to(addr);
                    continue;
                }
            }
            OpCode::JNE(reg, addr) => {
                if !get!(registers, reg).as_boolean()? {
                    execution.jump_to(addr);
                    continue;
                }
            }
            OpCode::HLT(return_register) => {
                return registers
                    .remove(return_register)
                    .map(|x| x.into())
                    .ok_or(Error::EmptyRegisters);
            }
            x => println!("{:?} is not implemented", x),
        }

        execution.next();
    }
}
